from __future__ import annotations
import datetime
import json
import logging
import threading
import time
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Callable, Optional, Union

import awscrt
import httpx
from awsiot import iotshadow, mqtt_connection_builder

from .model import JciHitachiAWSStatus, JciHitachiAWSStatusSupport

AWS_REGION = "ap-northeast-1"
AWS_COGNITO_IDP_ENDPOINT = f"cognito-idp.{AWS_REGION}.amazonaws.com"
AWS_COGNITO_ENDPOINT = f"cognito-identity.{AWS_REGION}.amazonaws.com"
AWS_COGNITO_CLIENT_ID = "7kfnjsb66ei1qt5s5gjv6j1lp6"
AWS_COGNITO_USERPOOL_ID = f"{AWS_REGION}_aTZeaievK"

#AMAZON_ROOT_CERT = os.path.join(os.path.dirname(os.path.abspath(__file__)), './cert/AmazonRootCA1.pem')
AWS_IOT_ENDPOINT = "iot-api.jci-hitachi-smarthome.com"
AWS_MQTT_ENDPOINT = f"a8kcu267h96in-ats.iot.{AWS_REGION}.amazonaws.com"

_LOGGER = logging.getLogger(__name__)

@dataclass
class AWSTokens:
    access_token: str
    id_token: str
    refresh_token: str
    expiration: float

@dataclass
class AWSIdentity:
    identity_id: str
    user_name: str
    user_attributes: dict

@dataclass
class JciHitachiMqttEvents:
    device_status: dict[str, JciHitachiAWSStatus] = field(default_factory=dict)
    device_support: dict[str, JciHitachiAWSStatusSupport] = field(default_factory=dict)
    device_control: dict[str, dict] = field(default_factory=dict)
    device_shadow: dict[str, dict] = field(default_factory=dict)
    mqtt_error: str = field(default_factory=str)
    device_status_event: threading.Event = field(default_factory=threading.Event)
    device_support_event: threading.Event = field(default_factory=threading.Event)
    device_control_event: threading.Event = field(default_factory=threading.Event)
    device_shadow_event: threading.Event = field(default_factory=threading.Event)
    mqtt_error_event: threading.Event = field(default_factory=threading.Event)


class JciHitachiAWSHttpConnection(ABC):
    """Abstract class for AWS http connections."""

    @abstractmethod
    def __init__(self, print_response: bool):
        self._print_response = print_response

    @abstractmethod
    def _handle_response(self, response: httpx.Response):
        ...

    @abstractmethod
    def _send(self):
        ...
    
    def get_data(self):
        raise NotImplementedError

    def maybe_print_http_response(self, response: httpx.Response) -> None:
        if not self._print_response:
            return
        
        print('===================================================')
        print(self.__class__.__name__, 'Response:')
        print('headers:', response.headers)
        print('status_code:', response.status_code)
        print('text:', json.dumps(response.json(), indent=True))
        print('===================================================')


class JciHitachiAWSCognitoConnection(JciHitachiAWSHttpConnection):
    """Connecting to Jci-Hitachi AWS Cognito API.

    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    aws_tokens : AWSTokens, optional
        If aws_tokens is given, it is used by request;
        otherwise, a login procedure is performed to obtain new aws_tokens,
        by default None.
    proxy : str, optional
        Proxy setting. Format:"IP:port", by default None. 
    print_response : bool, optional
        If set, all responses of httpx will be printed, by default False.
    """
    
    def __init__(self, 
        email: str,
        password: str,
        aws_tokens: Optional[AWSTokens] = None,
        proxy: Optional[str] = None,
        print_response: bool = False
    ):
        super().__init__(print_response)
        self._login_response = None
        self._email = email
        self._password = password
        self._proxies = {'http': proxy, 'https': proxy} if proxy else None

        if aws_tokens:
            self._aws_tokens = aws_tokens
        else:
            conn_status, self._aws_tokens = self.login()
            if conn_status != "OK":
                raise RuntimeError(f"An error occurred when signing into AWS Cognito Service: {conn_status}")
    
    def _generate_headers(self, target: str) -> dict[str, str]:
        normal_headers = {
            "X-Amz-Target": target,
            "User-Agent": "Dalvik/2.1.0",
            "content-type": "application/x-amz-json-1.1",
            "Accept" : "application/json",
        }
        return normal_headers

    def _handle_response(self, response: httpx.Response) -> tuple(str, dict):
        response_json = response.json()
        
        if response.status_code == httpx.codes.ok:    
            return "OK", response_json
        else:
            return f"{response_json['__type']} {response_json['message']}", response_json

    def _send(self, target: str, json_data : Optional[dict] = None):
        headers = self._generate_headers(target)

        req = httpx.post(
            f"https://{AWS_COGNITO_ENDPOINT if self.__class__.__name__ == 'GetCredentials' else AWS_COGNITO_IDP_ENDPOINT}/", 
            json=json_data,
            headers=headers,
            proxies=self._proxies,
        )

        self.maybe_print_http_response(req)

        return self._handle_response(req)

    @property
    def aws_tokens(self) -> AWSTokens:
        return self._aws_tokens

    def login(self, use_refresh_token : bool = False) -> tuple(str, AWSTokens):
        """Login API.

        Parameters
        ----------
        use_refresh_token : bool, optional
            Whether or not to use AWSTokens.refresh_token to login. 
            If AWSTokens is not provided, fallback to email and password, by default False

        Returns
        -------
        (str, AWSTokens)
            (status, aws tokens).
        """

        # https://docs.aws.amazon.com/cognito-user-identity-pools/latest/APIReference/API_InitiateAuth.html
        if use_refresh_token and self._aws_tokens != None:
            login_json_data = {
                "AuthFlow": 'REFRESH_TOKEN_AUTH',
                "AuthParameters":{
                    'REFRESH_TOKEN': self._aws_tokens.refresh_token,
                },
                "ClientId": AWS_COGNITO_CLIENT_ID,
            }
        else:
            use_refresh_token = False
            login_json_data = {
                "AuthFlow": 'USER_PASSWORD_AUTH',
                "AuthParameters": {
                    'USERNAME': self._email,
                    'PASSWORD': self._password,
                },
                "ClientId": AWS_COGNITO_CLIENT_ID,
            }

        login_headers = self._generate_headers("AWSCognitoIdentityProviderService.InitiateAuth")

        login_req = httpx.post(
            f"https://{AWS_COGNITO_IDP_ENDPOINT}", 
            json=login_json_data,
            headers=login_headers,
            proxies=self._proxies,
        )

        self.maybe_print_http_response(login_req)

        status, response = self._handle_response(login_req)

        aws_tokens = None
        if status == "OK":        
            auth_result = response["AuthenticationResult"]
            aws_tokens = AWSTokens(
                access_token = auth_result['AccessToken'],
                id_token = auth_result['IdToken'],
                refresh_token = self._aws_tokens.refresh_token if use_refresh_token else auth_result['RefreshToken'],
                expiration = time.time() + auth_result['ExpiresIn'],
            )
        return status, aws_tokens


class ChangePassword(JciHitachiAWSCognitoConnection):
    """API internal endpoint. 
    https://docs.aws.amazon.com/cognito-user-identity-pools/latest/APIReference/API_ChangePassword.html
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)
    
    def get_data(self, new_password):
        json_data = {
            "AccessToken": self._aws_tokens.access_token,
            "PreviousPassword": self._password,
            "ProposedPassword": new_password,
        }

        status, response = self._send("AWSCognitoIdentityProviderService.ChangePassword", json_data)

        return status, None


class GetUser(JciHitachiAWSCognitoConnection):
    """API internal endpoint. 
    https://docs.aws.amazon.com/cognito-user-identity-pools/latest/APIReference/API_GetUser.html
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)
    
    def get_data(self):
        json_data = {
            "AccessToken": self._aws_tokens.access_token,
        }

        status, response = self._send("AWSCognitoIdentityProviderService.GetUser", json_data)

        aws_identity = None
        if status == "OK":
            user_attributes = {attr["Name"]: attr["Value"] for attr in response["UserAttributes"]}
            aws_identity = AWSIdentity(
                identity_id = user_attributes["custom:cognito_identity_id"],
                user_name = response["Username"],
                user_attributes = user_attributes,
            )
        return status, aws_identity


class GetCredentials(JciHitachiAWSCognitoConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)

    def get_data(self, aws_identity):
        json_data = {
            "IdentityId": aws_identity.identity_id,
            "Logins": {
                f"{AWS_COGNITO_IDP_ENDPOINT}/{AWS_COGNITO_USERPOOL_ID}": self._aws_tokens.id_token,
            }
        }

        status, response = self._send("AWSCognitoIdentityService.GetCredentialsForIdentity", json_data)

        aws_credentials = None
        if status == "OK":
            aws_credentials = awscrt.auth.AwsCredentials(
                access_key_id = response["Credentials"]["AccessKeyId"],
                secret_access_key = response["Credentials"]["SecretKey"],
                session_token = response["Credentials"]["SessionToken"],
                expiration = datetime.datetime.fromtimestamp(response["Credentials"]["Expiration"]),
            )
        return status, aws_credentials


class JciHitachiAWSIoTConnection(JciHitachiAWSHttpConnection):
    """Connecting to Jci-Hitachi AWS IoT API.

    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    proxy : str, optional
        Proxy setting. Format:"IP:port", by default None. 
    print_response : bool, optional
        If set, all responses of httpx will be printed, by default False.
    """

    def __init__(self, aws_tokens, proxy=None, print_response=False):
        super().__init__(print_response)
        self._aws_tokens = aws_tokens
        self._proxies = {'http': proxy, 'https': proxy} if proxy else None
    
    def _generate_normal_headers(self) -> dict[str, str]:
        normal_headers = {
            "authorization": f"Bearer {self._aws_tokens.id_token}",
            "accesstoken": f"Bearer {self._aws_tokens.access_token}",
            "User-Agent": "Dalvik/2.1.0",
            "content-type" : "application/json",
            "Accept" : "application/json",
        }
        return normal_headers

    def _handle_response(self, response) -> tuple[int, str, dict]:
        response_json = response.json()
        if response.status_code == httpx.codes.ok:
            code = response_json["status"]["code"]
            if code == 0:
                return code, "OK", response_json
            elif code == 6:
                return code, "Invalid email or password", response_json
            elif code == 12:
                return code, "Invalid session token", response_json
            else:
                return code, "Unknown error", response_json
        else:
            return response.status_code, f"HTTP exception {response.status_code}", response_json

    def _send(self, api_name: str, json: Optional[dict] = None) -> tuple[str, dict]:
        req = httpx.post(
            f"https://{AWS_IOT_ENDPOINT}{api_name}",
            headers=self._generate_normal_headers(),
            json=json,
            proxies=self._proxies,
        )
        
        self.maybe_print_http_response(req)

        code, message, response_json = self._handle_response(req)

        return message, response_json


class GetAllDevice(JciHitachiAWSIoTConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens: AWSTokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self) -> tuple[str, dict]:
        return self._send("/GetAllDevice")


class GetAllGroup(JciHitachiAWSIoTConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens: AWSTokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self) -> tuple[str, dict]:
        return self._send("/GetAllGroup")


class GetAllRegion(JciHitachiAWSIoTConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens: AWSTokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self) -> tuple[str, dict]:
        return self._send("/GetAllRegion")


class GetAvailableAggregationMonthlyData(JciHitachiAWSIoTConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens: AWSTokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self, thing_name: str, time_start: int, time_end: int) -> tuple[str, dict]:
        json_data = {
            "ThingName": thing_name,
            "TimeStart": time_start,
            "TimeEnd": time_end,
        }
        return self._send("/GetAvailableAggregationMonthlyData", json_data)


class GetHistoryEventByUser(JciHitachiAWSIoTConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens: AWSTokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self, time_start: int, time_end: int) -> tuple[str, dict]:
        json_data = {
            "TimeStart": time_start,
            "TimeEnd": time_end,
        }
        return self._send("/GetHistoryEventByUser", json_data)


class ListSubUser(JciHitachiAWSIoTConnection):
    """API internal endpoint.
    
    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens: AWSTokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self) -> tuple[str, dict]:
        return self._send("/ListSubUser")


class JciHitachiAWSMqttConnection:
    """Connecting to Jci-Hitachi AWS MQTT to get latest events. 

    Parameters
    ----------
    get_credentials_callable : Callable
        Callable which takes no arguments and returns AwsCredentials.
    print_response : bool, optional
        If set, all responses of MQTT will be printed, by default False.
    """

    def __init__(self, get_credentials_callable: Callable, print_response: bool = False):
        self._get_credentials_callable = get_credentials_callable
        self._print_response = print_response
        
        self._mqttc = None
        self._shadow_mqttc = None
        self._client_tokens = {}
        self._mqtt_events = JciHitachiMqttEvents()

    def __del__(self):
        self.disconnect()

    @property
    def mqtt_events(self) -> JciHitachiMqttEvents:
        """MQTT events.

        Returns
        -------
        JciHitachiMqttEvents
            See JciHitachiMqttEvents.
        """

        return self._mqtt_events

    def _on_publish(self, topic: str, payload: bytes, dup, qos, retain, **kwargs):
        try:
            payload = json.loads(payload.decode())
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error(f"Mqtt topic {topic} published with payload {payload} cannot be decoded: {e}")
            return

        if self._print_response:
            print(f"Mqtt topic {topic} published with payload \n {payload}")

        split_topic = topic.split('/')

        if len(split_topic) >= 4 and split_topic[3] != "shadow":
            thing_name = split_topic[1]
            if split_topic[2] == "status" and split_topic[3] == "response":
                self._mqtt_events.device_status[thing_name] = JciHitachiAWSStatus(payload)
                self._mqtt_events.device_status_event.set()
            elif split_topic[2] == "registration" and split_topic[3] == "response":
                self._mqtt_events.device_support[thing_name] = JciHitachiAWSStatusSupport(payload)
                self._mqtt_events.device_support_event.set()
            elif split_topic[2] == "control" and split_topic[3] == "response":
                self._mqtt_events.device_control[thing_name] = payload
                self._mqtt_events.device_control_event.set()

    def _on_update_named_shadow_accepted(self, response):
        try:
            thing_name = self._client_tokens.pop(response.client_token)
        except:
            _LOGGER.error(f"An unknown shadow response is received. Client token: {response.client_token}")
            return
        
        if self._print_response:
            print(f"An `update` shadow response is received: {response.state.reported}")

        if response.state:
            if response.state.reported:
                self._mqtt_events.device_control[thing_name] = response.state.reported
                self._mqtt_events.device_control_event.set()
    
    def _on_update_named_shadow_rejected(self, error):
        _LOGGER.error(f"A shadow request {error.client_token} was rejected by the API: {error.code} {error.message}")

    def _on_get_named_shadow_accepted(self, response):
        try:
            thing_name = self._client_tokens.pop(response.client_token)
        except:
            _LOGGER.error(f"An unknown shadow response is received. Client token: {response.client_token}")
            return
        
        if self._print_response:
            print(f"A `get` shadow response is received: {response.state.reported}")

        if response.state:
            if response.state.reported:
                self._mqtt_events.device_shadow[thing_name] = response.state.reported
                self._mqtt_events.device_shadow_event.set()
    
    def _on_get_named_shadow_rejected(self, error):
        _LOGGER.error(f"A shadow request {error.client_token} was rejected by the API: {error.code} {error.message}")

    def _on_message(self, topic, payload, dup, qos, retain, **kwargs):
        return

    def disconnect(self) -> None:
        """Disconnect from the MQTT broker.
        """

        if self._mqttc is not None:
            self._mqttc.disconnect()

    def configure(self) -> None:
        """Configure MQTT."""

        cred_provider = awscrt.auth.AwsCredentialsProvider.new_delegate(self._get_credentials_callable)
        event_loop_group = awscrt.io.EventLoopGroup(1)
        host_resolver = awscrt.io.DefaultHostResolver(event_loop_group)
        client_bootstrap = awscrt.io.ClientBootstrap(event_loop_group, host_resolver)
        self._mqttc = mqtt_connection_builder.websockets_with_default_aws_signing(
            AWS_REGION,
            cred_provider,
            client_bootstrap=client_bootstrap,
            endpoint=AWS_MQTT_ENDPOINT,
            client_id=str(uuid.uuid4())
        )
        self._mqttc.on_message(self._on_message)
        self._shadow_mqttc = iotshadow.IotShadowClient(self._mqttc)

    def connect(
        self, 
        host_identity_id: str,
        shadow_names: Optional[Union[str, list[str]]] = None,
        thing_names: Optional[Union[str, list[str]]] = None
    ) -> bool:
        """Connect to the MQTT broker and start loop.
        
        Parameters
        ----------
        host_identity_id : str
            Host identity ID.
        shadow_names : str or list of str, optional
            Names to be subscribed in Shadow, by default None.
        thing_names : str or list of str, optional
            Things to be subscribed in Shadow, by default None.
        
        Returns
        -------
        bool
            A bool indicating whether the mqtt is successfully connected and subscribed.
        """
        
        try:
            connect_future = self._mqttc.connect()
            connect_future.result()
            _LOGGER.info("MQTT Connected.")
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error('MQTT connection failed with exception {}'.format(e))
            return False

        try:
            subscribe_future, _ = self._mqttc.subscribe(f"{host_identity_id}/#", awscrt.mqtt.QoS.AT_LEAST_ONCE, callback=self._on_publish)
            subscribe_future.result()
            
            if thing_names is not None and shadow_names is not None:
                shadow_names = [shadow_names] if isinstance(shadow_names, str) else shadow_names
                thing_names = [thing_names] if isinstance(thing_names, str) else thing_names
                
                for shadow_name in shadow_names:
                    for thing_name in thing_names:
                        update_accepted_subscribed_future, _ = self._shadow_mqttc.subscribe_to_update_named_shadow_accepted(
                            request=iotshadow.UpdateNamedShadowSubscriptionRequest(shadow_name=shadow_name, thing_name=thing_name),
                            qos=awscrt.mqtt.QoS.AT_LEAST_ONCE,
                            callback=self._on_update_named_shadow_accepted)

                        update_rejected_subscribed_future, _ = self._shadow_mqttc.subscribe_to_update_named_shadow_rejected(
                            request=iotshadow.UpdateNamedShadowSubscriptionRequest(shadow_name=shadow_name, thing_name=thing_name),
                            qos=awscrt.mqtt.QoS.AT_LEAST_ONCE,
                            callback=self._on_update_named_shadow_rejected)

                        # Wait for subscriptions to succeed
                        update_accepted_subscribed_future.result()
                        update_rejected_subscribed_future.result()

                        get_accepted_subscribed_future, _ = self._shadow_mqttc.subscribe_to_get_named_shadow_accepted(
                            request=iotshadow.GetNamedShadowSubscriptionRequest(shadow_name=shadow_name, thing_name=thing_name),
                            qos=awscrt.mqtt.QoS.AT_LEAST_ONCE,
                            callback=self._on_get_named_shadow_accepted)

                        get_rejected_subscribed_future, _ = self._shadow_mqttc.subscribe_to_get_named_shadow_rejected(
                            request=iotshadow.GetNamedShadowSubscriptionRequest(shadow_name=shadow_name, thing_name=thing_name),
                            qos=awscrt.mqtt.QoS.AT_LEAST_ONCE,
                            callback=self._on_get_named_shadow_rejected)

                        # Wait for subscriptions to succeed
                        get_accepted_subscribed_future.result()
                        get_rejected_subscribed_future.result()
        
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            self.disconnect()
            _LOGGER.error('MQTT subscription failed with exception {}'.format(e))
            return False
        return True

    def publish(self, topic: str, payload: dict) -> None:
        """Publish message.
        
        Parameters
        ----------
        topic : str
            Topic to publish.
        payload : dict
            Payload to publish.
        """

        try:
            publish_future, _ = self._mqttc.publish(topic, json.dumps(payload), awscrt.mqtt.QoS.AT_LEAST_ONCE)
            publish_future.result()
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error('Publish failed with exception: {}'.format(e))

    def publish_shadow(
        self,
        thing_name: str,
        command_name: str,
        payload: dict = {},
        shadow_name: Optional[str] = None
    ) -> None:
        """Publish message to IoT Shadow Service.
        
        Parameters
        ----------
        thing_name : str
            Thing name.
        command_name : str
            Command name, which can be `get` or `update`.
        payload : dict, optional
            Payload to publish, by default {}.
        shadow_name : str, optional
            Shadow name, by default None.
        """

        if command_name not in ["get", "update"]: # we don't subscribe delete
            raise ValueError("command_name must be one of `get` or `update`.")

        # The length of client token can't exceed 64 bytes, so we only use gateway mac address as the token.
        client_token = thing_name.split("_")[1] 
        self._client_tokens.update({client_token: thing_name})

        if shadow_name is None:
            if command_name == "get":
                publish_future = self._shadow_mqttc.publish_get_shadow(
                    iotshadow.GetShadowRequest(
                        client_token=client_token,
                        thing_name=thing_name
                    ),
                    qos=awscrt.mqtt.QoS.AT_LEAST_ONCE
                )
            elif command_name == "update":
                publish_future = self._shadow_mqttc.publish_update_shadow(
                    iotshadow.UpdateShadowRequest(
                        client_token=client_token,
                        state=iotshadow.ShadowState(reported=payload),
                        thing_name=thing_name
                    ),
                    qos=awscrt.mqtt.QoS.AT_LEAST_ONCE
                )
            elif command_name == "delete":
                publish_future = self._shadow_mqttc.publish_delete_shadow(
                    iotshadow.DeleteShadowRequest(
                        client_token=client_token,
                        thing_name=thing_name
                    ),
                    qos=awscrt.mqtt.QoS.AT_LEAST_ONCE
                )

        else:
            if command_name == "get":
                publish_future = self._shadow_mqttc.publish_get_named_shadow(
                    iotshadow.GetNamedShadowRequest(
                        client_token=client_token,
                        shadow_name=shadow_name,
                        thing_name=thing_name
                    ),
                    qos=awscrt.mqtt.QoS.AT_LEAST_ONCE
                )
            elif command_name == "update":
                publish_future = self._shadow_mqttc.publish_update_named_shadow(
                    iotshadow.UpdateNamedShadowRequest(
                        client_token=client_token,
                        shadow_name=shadow_name,
                        state=iotshadow.ShadowState(reported=payload),
                        thing_name=thing_name
                    ),
                    qos=awscrt.mqtt.QoS.AT_LEAST_ONCE
                )
            elif command_name == "delete":
                publish_future = self._shadow_mqttc.publish_delete_named_shadow(
                    iotshadow.DeleteNamedShadowRequest(
                        client_token=client_token,
                        shadow_name=shadow_name, 
                        thing_name=thing_name
                    ),
                    qos=awscrt.mqtt.QoS.AT_LEAST_ONCE
                )
        try:
            publish_future.result()
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error('Publish failed with exception: {}'.format(e))


class JciHitachiStatus:  # pragma: no cover
    idx = {}

    def __init__(self, status, default) -> None:
        self._status = status
        self._default = default

    @property
    def status(self):
        """All status.

        Returns
        -------
        dict
            All status.
        """

        return dict((key, getattr(self, key)) for key in self.idx)


class JciHitachiAC(JciHitachiStatus):  # pragma: no cover
    """Data class representing air conditioner status.

    Parameters
    ----------
    status : dict
        Status retrieved from JciHitachiStatusInterpreter.decode_status().
    default : int, optional
        Default value when a status doesn't exist, by default -1.
    """

    idx = {
        'power': 0,
        'mode': 1,
        'air_speed': 2,
        'target_temp': 3,
        'indoor_temp': 4,
        'sleep_timer': 6,
        'vertical_wind_swingable': 14,
        'vertical_wind_direction': 15,
        'horizontal_wind_direction': 17, 
        'mold_prev': 23,
        'fast_op': 26,
        'energy_save': 27,
        'sound_prompt': 30,
        'outdoor_temp': 33,
        'power_kwh': 40,
        'freeze_clean': 57,
    }

    def __init__(self, status, default=-1):
        super().__init__(status, default)
        
    @property
    def power(self):
        """Power. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx['power'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"
        elif v == 1:
            return "on"
        else:
            return "unknown"

    @property
    def mode(self):
        """Mode. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "cool", "dry", "fan", "auto", "heat", "unknown").
        """

        v = self._status.get(self.idx['mode'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "cool"
        elif v == 1:
            return "dry"
        elif v == 2:
            return "fan"
        elif v == 3:
            return "auto"
        elif v == 4:
            return "heat"
        else:
            return "unknown"

    @property
    def air_speed(self):
        """Air speed. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "auto", "silent", "low", "moderate", "high", "unknown").
        """

        v = self._status.get(self.idx['air_speed'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "auto"
        elif v == 1:
            return "silent"
        elif v == 2:
            return "low"
        elif v == 3:
            return "moderate"
        elif v == 4:
            return "high"
        else:
            return "unknown"

    @property
    def target_temp(self):
        """Target temperature. Controlable.

        Returns
        -------
        int
            Celsius temperature.
        """

        v = self._status.get(self.idx['target_temp'], self._default)
        return v

    @property
    def indoor_temp(self):
        """Indoor temperature.

        Returns
        -------
        int
            Celsius temperature.
        """

        v = self._status.get(self.idx['indoor_temp'], self._default)
        return v
    
    @property
    def max_temp(self):
        """Maximum target temperature.

        Returns
        -------
        int
            Celsius temperature.
        """

        return 32
    
    @property
    def min_temp(self):
        """Minimum target temperature.
        
        Returns
        -------
        int
            Celsius temperature.
        """

        return 16

    @property
    def sleep_timer(self):
        """Sleep timer. Controlable.
        
        Returns
        -------
        int
            Sleep timer (hours).
        """

        v = self._status.get(self.idx['sleep_timer'], self._default)
        return v
    
    @property
    def vertical_wind_swingable(self):
        """Vertical wind swingable. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """
        
        v = self._status.get(self.idx['vertical_wind_swingable'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "disabled"
        elif v == 1:
            return "enabled"
        else:
            return "unknown"

    @property
    def vertical_wind_direction(self):
        """Vertical wind direction. Controlable.

        Returns
        -------
        int
            Value between 0 to 15.
        """

        v = self._status.get(self.idx['vertical_wind_direction'], self._default)
        return v

    @property
    def horizontal_wind_direction(self):
        """Horizontal wind direction. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "auto", "leftmost", "middleleft", "central", "middleright", "rightmost", "unknown").
        """

        v = self._status.get(self.idx['horizontal_wind_direction'], self._default)
        
        if v > 0: v = 6-v

        if v == -1:
            return "unsupported"
        elif v == 0:
            return "auto"
        elif v == 1:
            return "leftmost"
        elif v == 2:
            return "middleleft"
        elif v == 3:
            return "central"
        elif v == 4:
            return "middleright"
        elif v == 5:
            return "rightmost"
        else:
            return "unknown"

    @property
    def mold_prev(self):
        """Mold prevention. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx['mold_prev'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "disabled"
        elif v == 1:
            return "enabled"
        else:
            return "unknown"
    
    @property
    def fast_op(self):
        """Fast operation. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx['fast_op'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "disabled"
        elif v == 1:
            return "enabled"
        else:
            return "unknown"
    
    @property
    def energy_save(self):
        """Energy saving. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx['energy_save'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "disabled"
        elif v == 1:
            return "enabled"
        else:
            return "unknown"

    @property
    def sound_prompt(self):
        """Sound prompt. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "enabled", "disabled", "unknown").
        """

        v = self._status.get(self.idx['sound_prompt'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "enabled"
        elif v == 1:
            return "disabled"
        else:
            return "unknown"

    @property
    def outdoor_temp(self):
        """Outdoor temperature.
        
        Returns
        -------
        int
            Celsius temperature.
        """

        v = self._status.get(self.idx['outdoor_temp'], self._default)
        return v
    
    @property
    def power_kwh(self):
        """Accumulated Kwh in a day.

        Returns
        -------
        float
            Kwh.
        """

        v = self._status.get(self.idx['power_kwh'], self._default)
        if v == -1:
            return v
        return v / 10.0

    @property
    def freeze_clean(self):
        """Freeze clean. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx['freeze_clean'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"
        elif v == 1:
            return "on"
        else:
            return "unknown"


class JciHitachiDH(JciHitachiStatus):  # pragma: no cover
    """Data class representing dehumidifier status.

    Parameters
    ----------
    status : dict
        Status retrieved from JciHitachiStatusInterpreter.decode_status().
    default : int, optional
        Default value when a status doesn't exist, by default -1.
    """

    idx = {
        'power': 0,
        'mode': 1,
        'target_humidity': 3,
        'indoor_humidity': 7,
        'wind_swingable': 8,
        'water_full_warning': 10,
        'clean_filter_notify': 11,
        'air_purify_level': 13,
        'air_speed': 14,
        'side_vent': 15,
        'sound_control': 16,
        'error_code': 18,
        'mold_prev': 19,
        'power_kwh': 29,
        'air_quality_value': 35,
        'air_quality_level': 36,
        'pm25_value': 37,
        'display_brightness': 39,
        'odor_level': 40,
        'air_cleaning_filter': 41
    }

    def __init__(self, status, default=-1):
        super().__init__(status, default)
    
    @property
    def power(self):
        """Power. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx['power'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"
        elif v == 1:
            return "on"
        else:
            return "unknown"
    
    @property
    def mode(self):
        """Mode. Controlable.

        Returns
        -------
        str
            One of (
            "unsupported", "auto", "custom", "continuous", "clothes_dry",
            "air_purify", "mold_prev", "low_humidity", "eco_comfort", "unknown"
            ).
        """

        v = self._status.get(self.idx['mode'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "auto"
        elif v == 1:
            return "custom"
        elif v == 2:
            return "continuous"
        elif v == 3:
            return "clothes_dry"
        elif v == 4:
            return "air_purify"
        elif v == 5:
            return "mold_prev"
        elif v == 8:
            return "low_humidity"
        elif v == 9:
            return "eco_comfort"
        else:
            return "unknown"

    @property
    def target_humidity(self):
        """Target humidity. Controlable.

        Returns
        -------
        int
            Relative humidity.
        """

        v = self._status.get(self.idx['target_humidity'], self._default)
        return v

    @property
    def indoor_humidity(self):
        """Indoor humidity.

        Returns
        -------
        int
            Relative humidity.
        """

        v = self._status.get(self.idx['indoor_humidity'], self._default)
        return v
    
    @property
    def max_humidity(self):
        """Maximum target humidity.

        Returns
        -------
        int
            Relative humidity.
        """

        return 70

    @property
    def min_humidity(self):
        """Minimum target humidity.

        Returns
        -------
        int
            Relative humidity.
        """

        return 40

    @property
    def wind_swingable(self):
        """Wind swingable. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx['wind_swingable'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"
        elif v == 1:
            return "on"
        else:
            return "unknown"

    @property
    def water_full_warning(self):
        """Water full warning.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx['water_full_warning'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"  # not activated
        elif v == 1:
            return "on"  # activated
        else:
            return "unknown"
    
    @property
    def clean_filter_notify(self):
        """Clean filter notify control. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx['clean_filter_notify'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "disabled"
        elif v == 1:
            return "enabled"
        else:
            return "unknown"

    @property
    def air_purify_level(self):
        """Air purify level. Not implemented.

        Returns
        -------
        str
            Not implemented.
        """
        return "unsupported"

    @property
    def air_speed(self):
        """Air speed. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "auto", "silent", "low", "moderate", "high", "unknown").
        """

        v = self._status.get(self.idx['air_speed'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "auto"
        elif v == 1:
            return "silent"
        elif v == 2:
            return "low"
        elif v == 3:
            return "moderate"
        elif v == 4:
            return "high"
        else:
            return "unknown"

    @property
    def side_vent(self):
        """Side vent.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx['side_vent'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"
        elif v == 1:
            return "on"
        else:
            return "unknown"

    @property
    def sound_control(self):
        """Sound control. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "silent", "button", "button+waterfull", "unknown").
        """

        v = self._status.get(self.idx['sound_control'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "silent"
        elif v == 1:
            return "button"
        elif v == 2:
            return "button+waterfull"
        else:
            return "unknown"

    @property
    def error_code(self):
        """Error code.

        Returns
        -------
        int
            Error code.
        """

        v = self._status.get(self.idx['error_code'], self._default)
        return v

    @property
    def mold_prev(self):
        """Mold prevention. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx['mold_prev'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"
        elif v == 1:
            return "on"
        else:
            return "unknown"

    @property
    def power_kwh(self):
        """Accumulated Kwh in a day.

        Returns
        -------
        float
            Kwh.
        """

        v = self._status.get(self.idx['power_kwh'], self._default)
        if v == -1:
            return v
        return v / 10.0

    @property
    def air_quality_value(self):
        """Air quality value. Not implemented.

        Returns
        -------
        int
            Not implemented.
        """
        return self._default

    @property
    def air_quality_level(self):
        """Air quality level. Not implemented.

        Returns
        -------
        str
            Not implemented.
        """
        return "unsupported"
    
    @property
    def pm25_value(self):
        """PM2.5 value.

        Returns
        -------
        int
            PM2.5 value.
        """

        v = self._status.get(self.idx['pm25_value'], self._default)
        return v
    
    @property
    def display_brightness(self):
        """Display brightness. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "bright", "dark", "off", "all_off" "unknown").
        """
        v = self._status.get(self.idx['display_brightness'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "bright"
        elif v == 1:
            return "dark"
        elif v == 2:
            return "off"
        elif v == 3:
            return "all_off"
        else:
            return "unknown"

    @property
    def odor_level(self):
        """Odor level.

        Returns
        -------
        str
            One of ("unsupported", "low", "middle", "high", "unknown").
        """
        v = self._status.get(self.idx['odor_level'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "low"
        elif v == 1:
            return "middle"
        elif v == 2:
            return "high"
        else:
            return "unknown"
    
    @property
    def air_cleaning_filter(self):
        """Air cleaning filter setting.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx['air_cleaning_filter'], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "disabled"
        elif v == 1:
            return "enabled"
        else:
            return "unknown"


class JciHitachiHE(JciHitachiStatus):  # pragma: no cover
    """Data class representing heat exchanger status. Not implemented.

    Parameters
    ----------
    status : dict
        Status retrieved from JciHitachiStatusInterpreter.decode_status().
    default : int, optional
        Default value when a status doesn't exist, by default -1.
    """

    idx = {}

    def __init__(self, status, default=-1):
        super().__init__(status, default)


class JciHitachiStatusSupport(JciHitachiStatus):  # pragma: no cover
    supported_type = {}

    def __init__(self, status, default=0):
        super().__init__(status, default)

    @property
    def brand(self):
        """Device brand.

        Returns
        -------
        str
            Device brand.
        """

        v = self._status.get(self.idx['brand'], self._default)
        return v

    @property
    def model(self):
        """Device model.

        Returns
        -------
        str
            Device model.
        """

        v = self._status.get(self.idx['model'], self._default)
        return v

    def _uni_v(self, v):
        return (((v >> 0x10) & 0xff) << 8) | (v >> 0x18)

    def _dual_v(self, v):
        low = (v >> 0x10) & 0xff
        high = (v >> 0x18) & 0xff
        return low, high

    def _functional_v(self, v):
        uni_v = self._uni_v(v)
        return ((uni_v >> i) & 0x1 for i in range(16))
    
    def limit(self, status_name, status_value):
        """Limit status_value within an acceptable range.

        Parameters
        ----------
        status_name : str
            Status name, which has to be in idx dict. E.g. JciHitachiAC.idx
        status_value : int
            Status value.

        Returns
        -------
        int or None
            If the status_value can be limited with an acceptable raneg, return int. 
            Otherwise, if the status_value is invalid, return None.
        """

        is_support, *v = getattr(self, status_name)
        if not is_support:
            return None
        
        supported_type = self.supported_type[status_name]
        if supported_type == "uni":
            return min(status_value, v[0])
        elif supported_type == "dual":
            return min(v[1], max(status_value, v[0]))
        elif supported_type == "functional":
            if v[status_value]: return status_value
        return None


class JciHitachiACSupport(JciHitachiStatusSupport):  # pragma: no cover
    """Data model representing supported air conditioner status.

    Parameters
    ----------
    status : dict
        Supported status retrieved from JciHitachiStatusInterpreter.decode_support().
    default : int, optional
        Default value when a status doesn't exist, by default 0.
    """

    idx = {
        'brand': "brand",
        'model': "model",
        'power': 0,
        'mode': 1,
        'air_speed': 2,
        'target_temp': 3,
        'indoor_temp': 4,
        'sleep_timer': 6,
        'vertical_wind_swingable': 14,
        'vertical_wind_direction': 15,
        'horizontal_wind_direction': 17,
        'mold_prev': 23,
        'fast_op': 26,
        'energy_save': 27,
        'sound_prompt': 30,
        'outdoor_temp': 33,
        'power_kwh': 40,
    }

    supported_type = {
        'brand': 'str',
        'model': 'str',
        'power': 'functional',
        'mode': 'functional',
        'air_speed': 'functional',
        'target_temp': 'dual',
        'indoor_temp': 'dual',
        'sleep_timer': 'uni',
        'vertical_wind_swingable': 'functional',
        'vertical_wind_direction': 'functional',
        'horizontal_wind_direction': 'functional',
        'mold_prev': 'functional',
        'fast_op': 'functional',
        'energy_save': 'functional',
        'sound_prompt': 'functional',
        'outdoor_temp': 'dual',
        'power_kwh': 'uni',
    }

    def __init__(self, status, default=0):
        super().__init__(status, default)

    @property
    def power(self):
        """Power. Controlable.

        Returns
        -------
        (bool, int, int)
            (is_support, off, on).
        """

        v = self._status.get(self.idx['power'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def mode(self):
        """Mode. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, (cool, dry, fan, auto, heat, 0...).
        """

        v = self._status.get(self.idx['mode'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def air_speed(self):
        """Air speed. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("auto", "silent", "low", "moderate", "high", 0...).
        """

        v = self._status.get(self.idx['air_speed'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def target_temp(self):
        """Target temperature. Controlable.

        Returns
        -------
        (bool, int, int)
            (is_support, minimum, maximum)
        """

        v = self._status.get(self.idx['target_temp'], self._default)
        supports = self._dual_v(v)

        return (v != 0, *supports)

    @property
    def indoor_temp(self):
        """Indoor temperature.

        Returns
        -------
        (bool, int, int)
            (is_support, minimum, maximum)
        """

        v = self._status.get(self.idx['indoor_temp'], self._default)
        supports = self._dual_v(v)

        return (v != 0, *supports)

    @property
    def sleep_timer(self):
        """Sleep timer. Controlable.
        
        Returns
        -------
        (bool, int)
            (is_support, maximum).
        """

        v = self._status.get(self.idx['sleep_timer'], self._default)
        support = self._uni_v(v)

        return (v != 0, support)
    
    @property
    def vertical_wind_swingable(self):
        """Vertical wind swingable. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("enabled", "disabled", 0...).
        """
        
        v = self._status.get(self.idx['vertical_wind_swingable'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def vertical_wind_direction(self):
        """Vertical wind direction. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("auto", "level1", "level2", "level3", "level4", "level5", "level6", "level7", 0...).
        """

        v = self._status.get(self.idx['vertical_wind_direction'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def horizontal_wind_direction(self):
        """Horizontal wind direction. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("auto", "leftmost", "middleleft", "central", "middleright", "rightmost", 0...).
        """

        v = self._status.get(self.idx['horizontal_wind_direction'], self._default)
        if v > 0: v = 6-v

        supports = self._functional_v(v)
        return (v != 0, *supports)

    @property
    def mold_prev(self):
        """Mold prevention. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("enabled", "disabled", 0...).
        """

        v = self._status.get(self.idx['mold_prev'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)
    
    @property
    def fast_op(self):
        """Fast operation. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("disabled", "enabled", 0...).
        """

        v = self._status.get(self.idx['fast_op'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)
    
    @property
    def energy_save(self):
        """Energy saving. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("enabled", "disabled", 0...).
        """

        v = self._status.get(self.idx['energy_save'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def sound_prompt(self):
        """Sound prompt. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("enabled", "disabled", 0...).
        """

        v = self._status.get(self.idx['sound_prompt'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def outdoor_temp(self):
        """Outdoor temperature.
        
        Returns
        -------
        (bool, int, int)
            (is_support, minimum, maximum)
        """

        v = self._status.get(self.idx['outdoor_temp'], self._default)
        supports = self._dual_v(v)

        return (v != 0, *supports)
    
    @property
    def power_kwh(self):
        """Accumulated Kwh in a day.

        Returns
        -------
        (bool, int)
            (is_support, maximum).
        """

        v = self._status.get(self.idx['power_kwh'], self._default)
        supports = self._uni_v(v)

        return (v != 0, supports)


class JciHitachiDHSupport(JciHitachiStatusSupport):  # pragma: no cover
    """Data model representing supported dehumidifier status.

    Parameters
    ----------
    status : dict
        Supported status retrieved from JciHitachiStatusInterpreter.decode_support().
    default : int, optional
        Default value when a status doesn't exist, by default 0.
    """

    idx = {
        'brand': "brand",
        'model': "model",
        'power': 0,
        'mode': 1,
        'target_humidity': 3,
        'indoor_humidity': 7,
        'wind_swingable': 8,
        'water_full_warning': 10,
        'clean_filter_notify': 11,
        'air_purify_level': 13,
        'air_speed': 14,
        'side_vent': 15,
        'sound_control': 16,
        'error_code': 18,
        'mold_prev': 19,
        'power_kwh': 29,
        'air_quality_value': 35,
        'air_quality_level': 36,
        'pm25_value': 37,
        'display_brightness': 39,
        'odor_level': 40,
        'air_cleaning_filter': 41
    }

    supported_type = {
        'brand': 'str',
        'model': 'str',
        'power': 'functional',
        'mode': 'functional',
        'target_humidity': 'dual',
        'indoor_humidity': 'dual',
        'wind_swingable': 'functional',
        'water_full_warning': 'functional',
        'clean_filter_notify': 'functional',
        'air_purify_level': 'functional', # not implemented
        'air_speed': 'functional',
        'side_vent': 'functional',
        'sound_control': 'functional',
        'error_code': 'uni',
        'mold_prev': 'functional',
        'power_kwh': 'uni',
        'air_quality_value': 'uni',
        'air_quality_level': 'functional',
        'pm25_value': 'uni',
        'display_brightness': 'functional',
        'odor_level': 'functional',
        'air_cleaning_filter': 'functional'
    }

    def __init__(self, status, default=0):
        super().__init__(status, default)
    
    @property
    def power(self):
        """Power. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            (is_support, off, on).
        """

        v = self._status.get(self.idx['power'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)
    
    @property
    def mode(self):
        """Mode. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, (auto, custom, continuous, clothes_dry, air_purify, mold_prev, air_supply, human_comfort, low_humidity, eco_comfort, 0...). 
        """

        v = self._status.get(self.idx['mode'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def target_humidity(self):
        """Target humidity. Controlable.

        Returns
        -------
        (bool, int, int)
            (is_support, minimum, maximum)
        """

        v = self._status.get(self.idx['target_humidity'], self._default)
        supports = self._dual_v(v)

        return (v != 0, *supports)

    @property
    def indoor_humidity(self):
        """Indoor humidity.

        Returns
        -------
        (bool, int, int)
            (is_support, minimum, maximum)
        """

        v = self._status.get(self.idx['indoor_humidity'], self._default)
        supports = self._dual_v(v)

        return (v != 0, *supports)

    @property
    def wind_swingable(self):
        """Wind swingable. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("off", "on", 0...).
        """
        

        v = self._status.get(self.idx['wind_swingable'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def water_full_warning(self):
        """Water full warning.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("off", "on", 0...).
        """

        v = self._status.get(self.idx['water_full_warning'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)
    
    @property
    def clean_filter_notify(self):
        """Clean filter notify control. Controlable.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx['clean_filter_notify'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def air_purify_level(self):
        """Air purify level. Not implemented.

        Returns
        -------
        str
            Not implemented.
        """
        
        return "unsupported"

    @property
    def air_speed(self):
        """Air speed. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("auto", "silent", "low", "moderate", "high", 0...).
        """

        v = self._status.get(self.idx['air_speed'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def side_vent(self):
        """Side vent.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("off", "on", 0...).
        """

        v = self._status.get(self.idx['side_vent'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def sound_control(self):
        """Sound control. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("silent", "button", "button+waterfull", 0...).    
        """

        v = self._status.get(self.idx['sound_control'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def error_code(self):
        """Error code.

        Returns
        -------
        (bool,)
            (is_support,).
        """

        v = self._status.get(self.idx['error_code'], self._default)

        return (v != 0,)

    @property
    def mold_prev(self):
        """Mold prevention. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("silent", "off", "on",, 0...).  
        """

        v = self._status.get(self.idx['mold_prev'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def power_kwh(self):
        """Accumulated Kwh in a day.

        Returns
        -------
        (bool, int)
            (is_support, maximum).
        """

        v = self._status.get(self.idx['power_kwh'], self._default)
        supports = self._uni_v(v)

        return (v != 0, supports)

    @property
    def air_quality_value(self):
        """Air quality value. Not implemented.

        Returns
        -------
        int
            Not implemented.
        """
        return self._default

    @property
    def air_quality_level(self):
        """Air quality level. Not implemented.

        Returns
        -------
        str
            Not implemented.
        """
        return "unsupported"
    
    @property
    def pm25_value(self):
        """PM2.5 value.

        Returns
        -------
        (bool, int)
            (is_support, maximum).
        """

        v = self._status.get(self.idx['pm25_value'], self._default)
        supports = self._uni_v(v)

        return (v != 0, supports)
    
    @property
    def display_brightness(self):
        """Display brightness. Controlable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("bright", "dark", "off", 0...).  
        """

        v = self._status.get(self.idx['display_brightness'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def odor_level(self):
        """Odor level.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("low", "middle", "high", 0...).  
        """

        v = self._status.get(self.idx['odor_level'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)
    
    @property
    def air_cleaning_filter(self):
        """Air cleaning filter setting.

        Returns
        -------
        (bool, Tuple[int]) 
            is_support, ("enabled", "disabled", 0...). status's reversed order.
        """

        v = self._status.get(self.idx['air_cleaning_filter'], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)


class JciHitachiHESupport(JciHitachiStatusSupport):  # pragma: no cover
    """Data model representing supported heat exchanger status. Not implemented.

    Parameters
    ----------
    status : dict
        Supported status retrieved from JciHitachiStatusInterpreter.decode_support().
    default : int, optional
        Default value when a status doesn't exist, by default 0.
    """

    idx = {}

    def __init__(self, status, default=0):
        super().__init__(status, default)


STATUS_DICT = {
    "AC": {
        'DeviceType': {
            "controlable": False,
            "is_numeric": False,
            "legacy_name": "DeviceType",
            "id2str": {
                1: "AC",
                2: "DH",
                3: "HE",
                4: "PM25_PANEL",
            }
        },
        'Switch': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": "power",
            "id2str": {
                0: "off",
                1: "on",
            }
        },
        'Mode': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": "mode",
            "id2str": {
                0: "cool",
                1: "dry",
                2: "fan",
                3: "auto",
                4: "heat",
            }
        },
        'FanSpeed': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": "air_speed",
            "id2str": {
                0: "auto",
                1: "silent",
                2: "low",
                3: "moderate",
                4: "high",
                5: "rapid",
                6: "express",
            }
        },
        'TemperatureSetting': {
            "controlable": True,
            "is_numeric": True,
            "legacy_name": "target_temp",
        },
        'IndoorTemperature': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": "indoor_temp",
        },
        'SleepModeRemainingTime': {
            "controlable": True,
            "is_numeric": True,
            "legacy_name": "sleep_timer",
        },
        'VerticalWindDirectionSwitch': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": "vertical_wind_swingable",
            "id2str": {
                0: "disabled",
                1: "enabled",
            }
        },
        'VerticalWindDirectionSetting': {
            "controlable": True,
            "is_numeric": True,
            "legacy_name": "vertical_wind_direction",
        },
        'HorizontalWindDirectionSetting': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": 'horizontal_wind_direction',
            "id2str": {
                0: "auto",
                1: "leftmost",
                2: "middleleft",
                3: "central",
                4: "middleright",
                5: "rightmost"
            }
        },
        'MildewProof': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": 'mold_prev',
            "id2str": {
                0: "disabled",
                1: "enabled",
            }
        },
        'QuickMode': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": 'fast_op',
            "id2str": {
                0: "disabled",
                1: "enabled",
            }
        },
        'PowerSaving': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name":'energy_save',
            "id2str": {
                0: "disabled",
                1: "enabled",
            }
        },
        'ControlTone': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name":'sound_prompt',
            "id2str": {
                0: "enabled",
                1: "disabled",
            }
        },
        'PowerConsumption': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name":'power_kwh',
        },
        'TaiseiaError': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        'FilterElapsedHour': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        'CleanSwitch': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name":'freeze_clean',
            "id2str": {
                0: "off",
                1: "on",
            }
        },
        'CleanNotification': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        'CleanStatus': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        'Error': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        'max_temp': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": 'max_temp',
        },
        'min_temp': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": 'min_temp',
        },
    },
    "DH": {
        'DeviceType': {
            "controlable": False,
            "is_numeric": False,
            "legacy_name": "DeviceType",
            "id2str": {
                1: "AC",
                2: "DH",
                3: "HE",
                4: "PM25_PANEL",
            }
        },
        'Switch': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": "power",
            "id2str": {
                0: "off",
                1: "on",
            }
        },
        'Mode': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": "mode",
            "id2str": {
                0: "auto",
                1: "custom",
                2: "continuous",
                3: "clothes_dry",
                4: "air_purify",
                5: "mold_prev",
                8: "low_humidity",
                9: "eco_comfort",
            }
        },
        'FanSpeed': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": "air_speed",
            "id2str": {
                0: "auto",
                1: "silent",
                2: "low",
                3: "moderate",
                4: "high",
            }
        },
        'MildewProof': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": "mold_prev",
            "id2str": {
                0: "disabled",
                1: "enabled",
            }
        },
        'ControlTone': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": 'sound_control',
            "id2str": {
                0: "silent",
                1: "button",
                2: "button+waterfull",
            }
        },
        'SaaControlTone': {  # currently not supported
            "controlable": False,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
            }
        },
        'PowerConsumption': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": 'power_kwh',
        },
        'Ion': {  # currently not supported
            "controlable": False,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
            }
        },
        'HumiditySetting': {
            "controlable": True,
            "is_numeric": True,
            "legacy_name": 'target_humidity',
        },
        'AutoWindDirection': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": 'wind_swingable',
            "id2str": {
                0: "disabled",
                1: "enabled",
            }
        },
        'KeypadLock': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
            }
        },
        'DisplayBrightness': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": 'display_brightness',
            "id2str": {
                0: "bright",
                1: "dark",
                2: "off",
                3: "all_off",
            }
        },
        'FilterControl': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": 'air_cleaning_filter',
            "id2str": {
                0: "disabled",
                1: "enabled",
            }
        },
        'PM25': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": 'pm25_value',
        },
        'IndoorHumidity': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": 'indoor_humidity',
        },
        'SideAirOutlet': {
            "controlable": False,
            "is_numeric": False,
            "legacy_name": 'side_vent',
            "id2str": {
                0: "off",
                1: "on",
            }
        },
        'Defrost': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
            }
        },
        'SmellIndex': {
            "controlable": False,
            "is_numeric": False,
            "legacy_name": 'odor_level',
            "id2str": {
                0: "low",
                1: "middle",
                2: "high",
            }
        },
        'CleanFilterNotification': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": 'clean_filter_notify',
            "id2str": {
                0: "disabled",
                1: "enabled",
            }
        },
        'TankFullNotification': {
            "controlable": False,
            "is_numeric": False,
            "legacy_name": 'water_full_warning',
            "id2str": {
                0: "off",  # not activated
                1: "on",  # activated
            }
        },
        'TaiseiaError': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        'Error': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": 'error_code',
        },
        'max_humidity': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": 'max_humidity',
        },
        'min_humidity': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": 'min_humidity',
        },
    },
    "HE": {
        'DeviceType': {
            "controlable": False,
            "is_numeric": False,
            "legacy_name": "DeviceType",
            "id2str": {
                1: "AC",
                2: "DH",
                3: "HE",
                4: "PM25_PANEL",
            }
        },
        'Switch': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "off",
                1: "on",
            }
        },
        'Mode': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "air_condition",
                1: "dehumidification",
                2: "air_supply",
                3: "auto",
                4: "heater",
            }
        },
        'FanSpeed': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "auto",
                1: "silent",
                2: "low",
                3: "moderate",
                4: "high",
            }
        },
        'IndoorTemperature': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        'TaiseiaError': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        'CleanFilterNotification': {
            "controlable": False,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "disabled",
                1: "enabled",
            }
        },
        'BreathMode': {
            "controlable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "auto",
                1: "energy_recovery",
                2: "normal"
            }
        },
        'FrontFilterNotification': {
            "controlable": False,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "disabled",
                1: "enabled",
            }
        },
        'Pm25FilterNotification': {
            "controlable": False,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "disabled",
                1: "enabled",
            }
        },
        'Error': {
            "controlable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
    },
    "PM25_PANEL": {

    }
}

class JciHitachiAWSStatus:
    """Data class representing `AWSThing` status.

    Parameters
    ----------
    raw_status : dict
        Status retrieved from `JciHitachiAWSMqttConnection` _on_publish() callback.
    legacy : bool, optional
        Whether the raw_status is a legacy status, i.e. derived from subclasses of `JciHitachiStatus`, by default False.
    """

    device_type_mapping = {
        1: "AC",
        2: "DH",
        3: "HE",
        4: "PM25_PANEL",
    }

    def __init__(self, raw_status: dict, legacy=False) -> None:
        self._raw_status: dict = raw_status
        self._status: dict = raw_status if legacy else self._preprocess(raw_status)
        self._device_type: str = self._status["DeviceType"]

    def __getattr__(self, name):
        return self._status.get(name, "unsupported")
    
    def __repr__(self) -> str:
        return str(self._status)

    def _preprocess(self, raw_status):
        # device type
        if "DeviceType" not in raw_status or \
            ("DeviceType" in raw_status and raw_status["DeviceType"] not in self.device_type_mapping):
            raise AttributeError("`DeviceType` isn't in the raw status or has an invalid value.")
        
        if "PowerConsumption" in raw_status:
            raw_status["PowerConsumption"] /= 10.0

        status = {}
        device_type = self.device_type_mapping[raw_status["DeviceType"]]
        for key, value in raw_status.items():
            if key in STATUS_DICT[device_type]:
                if STATUS_DICT[device_type][key]["is_numeric"]:
                    status[key] = value
                else:
                    status[key] = STATUS_DICT[device_type][key]["id2str"].get(value, "unknown")

        return status

    @property
    def status(self):
        """All status.

        Returns
        -------
        dict
            All status.
        """

        return self._status
    
    @property
    def legacy_status(self):
        """All legacy status name used by the old API.

        Returns
        -------
        JciHitachiAWSStatus
            Status with legacy name.
        """

        status = {}
        device_type = self._status["DeviceType"]
        for status_name, status_value in self._status.items():
            if status_name in STATUS_DICT[device_type]:
                if STATUS_DICT[device_type][status_name]["legacy_name"] is None:
                    # no legacy name
                    status.update({status_name: status_value})
                else:
                    status.update({STATUS_DICT[device_type][status_name]["legacy_name"]: status_value})
        return JciHitachiAWSStatus(status, legacy=True)

    @staticmethod
    def str2id(device_type: str, status_name: str, status_value: int = None, status_str_value: str = None, support_code: int = None):
        is_valid = (status_value is not None) ^ (status_str_value is not None)

        # Name check
        if is_valid:
            if status_name not in STATUS_DICT[device_type]:
                legacy2new = {specs["legacy_name"]: new_status_name for new_status_name, specs in STATUS_DICT[device_type].items()}
                if status_name in legacy2new:
                    status_name = legacy2new[status_name]
                else:
                    is_valid = False

        # Value check
        if is_valid:
            if status_str_value is not None:
                str2id_dict = {value:key for key, value in STATUS_DICT[device_type][status_name]["id2str"].items()}
                if status_str_value in str2id_dict:
                    status_value = str2id_dict[status_str_value]
                else:
                    is_valid = False
            else:
                if not STATUS_DICT[device_type][status_name]["is_numeric"] and status_value not in STATUS_DICT[device_type][status_name]["id2str"]:
                    is_valid = False
        
        # if support_code is specified, we check whether the given status value is valid for the device
        if is_valid and support_code is not None: 
            if STATUS_DICT[device_type][status_name]["is_numeric"]:
                if status_value > support_code.status[status_name]:
                    is_valid = False
            else:
                if 2 ** status_value & support_code.status[status_name] == 0:
                    is_valid = False

        return is_valid, status_name, status_value

    def set_new_status(self, name: str, value: int):
        if STATUS_DICT[self._device_type][name]["is_numeric"]:
            self._status[name] = value
        else:
            self._status[name] = STATUS_DICT[self._device_type][name]["id2str"].get(value, "unknown")


class JciHitachiAWSStatusSupport:
    """Data class representing `AWSThing` status support.

    Parameters
    ----------
    raw_status : dict
        Status retrieved from `JciHitachiAWSMqttConnection` _on_publish() callback.
    """

    extended_mapping = {
        "FirmwareId": None,
        "Model": "model",
        "Brand": "brand",
        "FindMe": None,
    }

    device_type_mapping = JciHitachiAWSStatus.device_type_mapping

    def __init__(self, raw_status: dict) -> None:
        self._raw_status: dict = raw_status
        self._status: dict = self._preprocess(raw_status)

    def __getattr__(self, name):
        return self._status.get(name, "unsupported")
    
    def __repr__(self) -> str:
        return str(self._status)

    def _preprocess(self, status):
        status = status.copy()
        # device type
        status["DeviceType"] = self.device_type_mapping[status["DeviceType"]]
        
        status["Brand"] = "HITACHI"
        
        if status["DeviceType"] == "AC":
            status["max_temp"] = status["TemperatureSetting"] & 255 if "TemperatureSetting" in status else 32
            status["min_temp"] = status["TemperatureSetting"] >> 8 & 255 if "TemperatureSetting" in status else 16
        elif status["DeviceType"] == "DH":
            status["max_humidity"] = status["HumiditySetting"] & 255 if "HumiditySetting" in status else 70
            status["min_humidity"] = status["HumiditySetting"] >> 8 & 255 if "HumiditySetting" in status else 40

        return status

    @property
    def status(self):
        """All status.

        Returns
        -------
        dict
            All status.
        """

        return self._status


import json
import os
import ssl

import httpx

from .utility import convert_hash

API_ENDPOINT = "https://api.jci-hitachi-smarthome.com/3.6/"
API_ID = "ODM-HITACHI-APP-168d7d31bbd2b7cbbd"
API_KEY = "23f26d38921dda92c1c2939e733bca5e"
API_SSL_CERT = os.path.join(os.path.dirname(os.path.abspath(__file__)), "cert/api-jci-hitachi-smarthome-com-chain.pem")
API_SSL_CONTEXT = httpx.create_ssl_context()
API_SSL_CONTEXT.load_verify_locations(cafile=API_SSL_CERT)
API_SSL_CONTEXT.set_ciphers("DEFAULT@SECLEVEL=1")  # the cert uses SHA1-RSA1024bits ciphers so unfortunately we have to lower the security level
API_SSL_CONTEXT.hostname_checks_common_name=True  # the cert lacks a subjectaltname


APP_PLATFORM = 2  # 1=IOS 2=Android
APP_VERSION = "10.20.900"


class JciHitachiConnection:  # pragma: no cover
    """Connecting to Jci-Hitachi API to get data or send commands.

    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    session_token : str, optional
        If session_token is given, it is used by request, 
        otherwise perform a login procedure to obtain a new token,
        by default None.
    proxy : str, optional
        Proxy setting. Format:"IP:port", by default None. 
    print_response : bool, optional
        If set, all responses of httpx will be printed, by default False.
    """
    
    def __init__(self, email, password, session_token=None, proxy=None, print_response=False):
        self._login_response = None
        self._email = email
        self._password = password
        self._print_response = print_response
        self._proxies = {'http': proxy, 'https': proxy} if proxy else None

        if session_token:
            self._session_token = session_token
        else:
            self._session_token = ""
            self.login()
    
    @property
    def logged(self):
        return True if self._session_token is not None else False

    @property
    def session_token(self):
        return self._session_token

    def _generate_normal_headers(self):
        normal_headers = {
            "X-DK-Application-Id" : API_ID,
            "X-DK-API-Key" : API_KEY,
            "X-DK-Session-Token" : self._session_token, 
            "User-Agent": "Dalvik/2.1.0",
            "content-type" : "application/json",
            "Accept" : "application/json",
        }
        return normal_headers

    def _handle_response(self, response):
        response_json = response.json()
        code = response_json["status"]["code"]
        if code == 0:
            return code, "OK", response_json
        elif code == 6:
            return code, "Invalid email or password", response_json
        elif code == 12:
            return code, "Invalid session token", response_json
        else:
            return code, "Unknown error", response_json

    def _send(self, api_name, json=None):
        code = 12
        while code == 12:
            req = httpx.post(
                "{}{}".format(API_ENDPOINT, api_name),
                headers=self._generate_normal_headers(),
                json=json,
                verify=API_SSL_CONTEXT,
                proxies=self._proxies,
            )
            if self._print_response:
                self.print_response(req)

            code, message, response_json = self._handle_response(req)
            
            if code == 12:
                if not self.login():
                    break

        return message, response_json

    def login(self):
        login_json_data = {
            "ServerLogin": {
                "Email": self._email,
                "HashCode": f"{self._email}{convert_hash(f'{self._email}{self._password}')}"
            },
            "AppVersion": {
                "Platform": int(APP_PLATFORM),
                "Version": APP_VERSION
            }
        }

        login_headers = {
            "X-DK-Application-Id" : API_ID,
            "X-DK-API-Key" : API_KEY,
            "User-Agent": "Dalvik/2.1.0",
            "content-type" : "application/json",
            "Accept" : "application/json",
        }

        login_req = httpx.post("{}{}".format(API_ENDPOINT, "UserLogin.php"), 
            json=login_json_data,
            headers=login_headers,
            verify=API_SSL_CONTEXT,
            proxies=self._proxies,
        )

        if self._print_response:
            self.print_response(login_req)

        if login_req.status_code == httpx.codes.ok:
            code, message, self._login_response = self._handle_response(login_req)
            if message == "OK":
                self._session_token = self._login_response["results"]["sessionToken"]
                return True
                
        return False

    def get_data(self):
        raise NotImplementedError

    def print_response(self, response):
        print('===================================================')
        print(self.__class__.__name__, 'Response:')
        print('headers:', response.headers)
        print('status_code:', response.status_code)
        print('text:', json.dumps(response.json(), indent=True))
        print('===================================================')


class RegisterMobileDevice(JciHitachiConnection):  # pragma: no cover
    """API internal endpoint. (Unused)
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)

    def get_data(self, device_token):
        json_data = {
            "DeviceType": APP_PLATFORM,
            "DeviceToken": device_token
        }

        return self._send("RegisterMobileDevice.php", json_data)


class UpdateUserCredential(JciHitachiConnection):  # pragma: no cover
    """API internal endpoint. (Tested)
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)

    def get_data(self, new_password):
        """Get data from the endpoint.

        Parameters
        ----------
        new_password : str
            New password.

        Returns
        -------
        (str, dict)
            (message, response_json)
        """

        json_data = {
            "ServerLogin": {
                "OldEmail": self._email,
                "OldPassword": convert_hash(f'{self._email}{self._password}'),
                "NewPassword": convert_hash(f'{self._email}{new_password}'),
            }
        }

        return self._send("UpdateUserCredential.php", json_data)


class GetServerLastUpdateInfo(JciHitachiConnection):  # pragma: no cover
    """API internal endpoint. (Unused)
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)

    def get_data(self):
        """Get data from the endpoint.

        Returns
        -------
        (str, dict)
            (message, response_json)
        """

        return self._send("GetServerLastUpdateInfo.php")


class GetPeripheralsByUser(JciHitachiConnection):  # pragma: no cover
    """API internal endpoint. (Tested)
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)
    
    def get_data(self):
        """Get data from the endpoint.

        Returns
        -------
        (str, dict)
            (message, response_json)
        """

        return self._send("GetPeripheralsByUser.php")


class GetDataContainerByID(JciHitachiConnection):  # pragma: no cover
    """API internal endpoint. (Tested)
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)
    
    def get_data(self, picked_peripheral_json):
        """Get data from the endpoint.

        Parameters
        ----------
        picked_peripheral_json : dict
            Picked peripheral_json.

        Returns
        -------
        (str, dict)
            (message, response_json)
        """

        ContMID = picked_peripheral_json["Peripherals"][0]["DataContainer"][0]["ContMID"]
        ContDID_1 = picked_peripheral_json["Peripherals"][0]["DataContainer"][0]["ContDetails"][0]["ContDID"]
        ContDID_2 = picked_peripheral_json["Peripherals"][0]["DataContainer"][0]["ContDetails"][1]["ContDID"]

        json_data = {
            "Format": 0,
            "DataContainer": [
                {"ContMID" : ContMID,
                 "ContDetails":   
                    [
                        {"ContDID": ContDID_1},
                        {"ContDID": ContDID_2}
                    ]
                }       
            ]
        }

        return self._send("GetDataContainerByID.php", json_data)


class GetPeripheralByGMACAddress(JciHitachiConnection):  # pragma: no cover
    """API internal endpoint. (Unused)
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)
    
    def get_data(self, peripheral_json):
        """Get data from the endpoint.

        Parameters
        ----------
        peripheral_json : dict
            peripheral_json.

        Returns
        -------
        (str, dict)
            (message, response_json)
        """

        GMACAddress = peripheral_json["results"][0]["GMACAddress"]
        json_data = {
            "Data" : [
                {"GMACAddress": GMACAddress}
            ]
        }

        return self._send("GetPeripheralByGMACAddress.php", json_data)


class CreateJob(JciHitachiConnection):  # pragma: no cover
    """API internal endpoint. (Tested)
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)
    
    def get_data(self, gateway_id, device_id, task_id, job_info):
        """Get data from the endpoint.

        Parameters
        ----------
        gateway_id : int
            Peripheral.gateway_id.
        device_id : int
            Random device ID.
        task_id : int
            Task ID (serial number).
        job_info : str
            Base64 job info.

        Returns
        -------
        (str, dict)
            (message, response_json)
        """

        json_data = {
            "Data": [
                {
                    "GatewayID": gateway_id,
                    "DeviceID": device_id, # random device ID
                    "TaskID": task_id, # serial number started from 1
                    "JobInformation": job_info
                }
            ]
        }

        return self._send("CreateJob.php", json_data)


class GetJobDoneReport(JciHitachiConnection):  # pragma: no cover
    """API internal endpoint. (Tested)
    
    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)

    def get_data(self, device_id):
        """Get data from the endpoint.

        Parameters
        ----------
        device_id : int
            Random device ID.

        Returns
        -------
        (str, dict)
            (message, response_json)
        """
        
        json_data = {
            "DeviceID": device_id
        }

        return self._send("GetJobDoneReport.php", json_data)



import setuptools

from JciHitachi import __author__, __version__

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

install_requires = [
    "awsiotsdk==1.12.0", # See: https://github.com/pypa/pip/issues/5898
    "httpx",
    "paho-mqtt",
]
tests_require = [
    "pytest>=6.2",
    "pytest-cov",
]


if __name__ == "__main__":
    setuptools.setup(
        name="LibJciHitachi",
        version=__version__,
        author=__author__,
        author_email="qqaatw@gmail.com",
        description="A library for controlling Jci Hitachi devices.",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/qqaatw/LibJciHitachi",
        project_urls={
            "Issue Tracker": "https://github.com/qqaatw/LibJciHitachi/issues",
            "Documentation": "https://libjcihitachi.readthedocs.io/en/latest/",
        },
        classifiers=[
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
        ],
        packages=setuptools.find_packages(include=['JciHitachi']),
        package_data={'JciHitachi': ['cert/*.pem']},
        python_requires=">=3.7",
        install_requires=install_requires,
        tests_require=tests_require,
    )
